// This file is part of LeebA, a lightweight C++ template library
// for convex optimization and control.
//
// Copyright (C) 2008 Timothee Hunter <tjhunter@stanford.edu>
//
// LeebA is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 3 of the License, or (at your option) any later version.
//
// LeebA is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License  for
// more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with LeebA in the file COPYING.LESSER at the root of the
// LeebA package. If not, see <http://www.gnu.org/licenses/>.
/*
 * Monom.h
 *
 *  Created on: Dec 19, 2008
 *      Author: Timothy Hunter <tjhunter@stanford.edu>
 */

#include <iostream>
#include <boost/functional/hash.hpp>
#include <tr1/unordered_map>
#include "Symbol.h"

#ifndef MONOM_H_
#define MONOM_H_

namespace Eigen
{

/**
 * @brief Representation of a monom (product of symbols)
 * The current implementation uses a hash map to map symbols in the monom to the exponent index.
 */
class Monom
{
private:
  typedef std::tr1::unordered_map<Eigen::Symbol,int,boost::hash<Symbol> > MapType;
  static const MapType::size_type size_ = 10;
public:
  Monom() : map_(size_), hash_(0) {}

  Monom(const Symbol & s) :
    map_(size_)
  {
    (*this)*=s;
  }

  //FIXME integrate it with Eigen's output classes.
  friend std::ostream & operator<<(std::ostream &os, const Monom & m)
  {
    os<<"(";
    for(MapType::const_iterator it=m.map_.begin();it!=m.map_.end();++it)
    {
      os<<(it->first);
      if(it->second!=1)
        os<<"^("<<(it->second)<<")";
      else
      {
        if((++MapType::const_iterator(it))!=m.map_.end())
        os<<'.';

      }

    }
    os<<")";
    return os;
  }

  Monom & operator*=(const Symbol & s)
  {
    add_key(s,1);
    compute_hash();
    return *this;
  }

  Monom & operator*=(const Monom & other)
  {
    for(MapType::const_iterator it=other.map_.begin();it!=other.map_.end();++it)
      add_key(it->first,it->second);
    compute_hash();
    return *this;
  }

  inline friend std::size_t hash_value(const Monom & m)
  {
    return m.hash_;
  }


  inline friend bool operator==(const Monom & m1, const Monom & m2)
  {
    if(m1.hash_!=m2.hash_)
      return false;
    MapType::const_iterator it1=m1.map_.begin();
    MapType::const_iterator it2=m2.map_.begin();
    while(it1!=m1.map_.end()&&it2!=m2.map_.end())
    {
      if(it1==m1.map_.end()||it2==m2.map_.end())
        return false;
      if(!(it1->first==it2->first))
        return false;
      if(it1->second!=it2->second)
        return false;
      ++it1;
      ++it2;
    }
    return true;
  }

  inline friend bool operator!=(const Monom & m1, const Monom & m2)
  {
    return !(m1==m2);
  }

private:

  void add_key(const Symbol & s, int factor=1)
  {
    if(map_.find(s)==map_.end())
    {
      map_[s]=factor;
      return;
    }
    map_[s]+=factor;
    if(map_[s]==0)
      map_.erase(s);
  }

  void compute_hash()
  {
    hash_=0;
    for(MapType::const_iterator it=map_.begin();it!=map_.end();++it)
    {
      boost::hash_combine(hash_,it->first);
      boost::hash_combine(hash_,it->second);
    }
  }

  MapType map_;

  std::size_t hash_;
};

inline Monom operator*(const Symbol & s1, const Symbol & s2)
{
  return Monom(s1)*=s2;
}

inline Monom operator*(const Monom & m, const Symbol & s)
{
  return Monom(m)*=s;
}

inline Monom operator*(const Symbol & s, const Monom & m)
{
  return Monom(m)*=s;
}

inline Monom operator*(const Monom & m1, const Monom & m2)
{
  return Monom(m1)*=m2;
}

}


#endif /* MONOM_H_ */
