// This file is part of LeebA, a lightweight C++ template library
// for convex optimization and control.
//
// Copyright (C) 2008 Timothee Hunter <tjhunter@stanford.edu>
//
// LeebA is free software; you can redistribute it and/or
// modify it under the terms of the GNU Lesser General Public
// License as published by the Free Software Foundation; either
// version 3 of the License, or (at your option) any later version.
//
// LeebA is distributed in the hope that it will be useful, but WITHOUT ANY
// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
// FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License  for
// more details.
//
// You should have received a copy of the GNU Lesser General Public
// License along with LeebA in the file COPYING.LESSER at the root of the
// LeebA package. If not, see <http://www.gnu.org/licenses/>.

/*
 * Multinom.h
 *
 *  Created on: Dec 19, 2008
 *      Author: Timothy Hunter <tjhunter@stanford.edu>
 */

#include "Symbol.h"
#include "Monom.h"

#ifndef MULTINOM_H_
#define MULTINOM_H_
namespace Eigen
{

/** @brief Representation of a multinom (linear combination of monoms)
 *
 * The implementation is based on hash tables. It is extremely slow when creating a new element, but easy to implement.
 *
 * @note The evaluation is not implemented yet.
 *
 */
template<typename T>
class Multinom
{

private:
  typedef std::tr1::unordered_map<Monom,T,boost::hash<Monom> > MapType;
  static const typename MapType::size_type size_ = 5; //Default size of the hash table

public:
  Multinom(const T & t = T(0));

  Multinom(const Symbol & s);

  Multinom(const Monom &, const T & t=T(1));

  Multinom & operator*=(const T &);

  Multinom & operator*=(const Monom &);

  Multinom & operator*=(const Multinom &);

  Multinom & operator+=(const Multinom &);

  Multinom & operator-=(const Multinom &);

  friend bool operator==(const Multinom & m1, const Multinom & m2)
  {
    if(m1.hash_!=m2.hash_)
      return false;
    typename MapType::const_iterator it1=m1.map_.begin();
    typename MapType::const_iterator it2=m2.map_.begin();
    while(it1!=m1.map_.end()&&it2!=m2.map_.end())
    {
      if(it1==m1.map_.end()||it2==m2.map_.end())
        return false;
      if(!(it1->first==it2->first))
        return false;
      if(it1->second!=it2->second)
        return false;
      ++it1;
      ++it2;
    }
    return true;
  }

  inline friend std::size_t hash_value(const Multinom<T> & m)
  {
    return m.hash_;
  }

  //FIXME integrate the output with the classes used by Eigen.
  friend std::ostream & operator<<(std::ostream &os, const Multinom<T> &m)
  {
    typedef typename Multinom<T>::MapType::const_iterator const_iterator;
    os<<"[";
    for(const_iterator it=m.map_.begin();it!=m.map_.end();++it)
      {
//        if(!(it->second==T(1)))
          os<<"("<<it->second<<")";
        os<<it->first;
        const_iterator it2=it;
        it2++;
        if(it2!=m.map_.end())
          os<<"+";
      }
    return os<<"]";
  }

private:
  static void product_add(const MapType & map1, MapType & map2, const Monom & m, const T & t);

  void compute_hash();

  MapType map_;
  std::size_t hash_;
};

template<typename T>
struct NumTraits<Multinom<T> >
{
  typedef Multinom<typename NumTraits<T>::Real> Real;
  typedef Multinom<T> FloatingPoint;
  enum {
    IsComplex = NumTraits<T>::IsComplex,
    HasFloatingPoint = NumTraits<T>::HasFloatingPoint,
    ReadCost = 20,
    AddCost = 40,
    MulCost = 20
  };
};

template<typename T>
Multinom<T>::Multinom(const T &t) :
  map_(size_)
{
  map_[Monom()]=t;
  compute_hash();
}

template<typename T>
Multinom<T>::Multinom(const Symbol &s) :
  map_(size_)
{
  map_[Monom()]=T(0);
  map_[Monom(s)]=T(1);
  compute_hash();
}

template<typename T>
Multinom<T>::Multinom(const Monom &m, const T & t) :
  map_(size_)
{
  map_[m]=t;
  compute_hash();
}

template<typename T>
Multinom<T> & Multinom<T>::operator*=(const T &t)
{
  if(t==T(0))
    return (*this=Multinom<T>());
  typedef typename Multinom<T>::MapType::iterator iterator;
  for(iterator it=map_.begin();it!=map_.end();++it)
    (it->second)*=t;
  compute_hash();
  return *this;
}

template<typename T>
Multinom<T> & Multinom<T>::operator*=(const Monom &m)
{
  MapType map(map_.size());
  product_add(map_,map,m,T(1));
  map_=map;
  compute_hash();
  return *this;
}

template<typename T>
Multinom<T> & Multinom<T>::operator*=(const Multinom<T> &m)
{
  MapType map(map_.size());

  typedef typename Multinom<T>::MapType::const_iterator const_iterator;
  for(const_iterator it=m.map_.begin();it!=m.map_.end();++it)
      product_add(map_,map,it->first,it->second);
  map_=map;
  compute_hash();
  return *this;
}

template<typename T>
Multinom<T> & Multinom<T>::operator+=(const Multinom<T> &m)
{
  typedef typename Multinom<T>::MapType::const_iterator iterator;
  for(iterator it=m.map_.begin();it!=m.map_.end();++it)
  {
    if(map_.find(it->first)==map_.end())
      map_[it->first]=it->second;
    else
    {
      map_[it->first]+=it->second;
      if(map_[it->first]==T(0) && it->first != Monom())
        map_.erase(it->first);
    }
  }
  compute_hash();
  return *this;
}


template<typename T>
void Multinom<T>::product_add(const typename Multinom<T>::MapType & map1, typename Multinom<T>::MapType & map2, const Monom & m, const T & t)
{
  typedef typename Multinom<T>::MapType::const_iterator const_iterator;
  for(const_iterator it=map1.begin();it!=map1.end();++it)
  {
    const T & tProd = t*(it->second);
    const Monom & mProd = (it->first) * m;
    if(map2.find(mProd) == map2.end() )
      map2[mProd] = tProd;
    else map2[mProd] += tProd;
    if(map2[mProd]==T(0) && mProd != Monom())
      map2.erase(mProd);
  }
}

template<typename T>
Multinom<T> & Multinom<T>::operator-=(const Multinom<T> &m)
{
  typedef typename Multinom<T>::MapType::const_iterator iterator;
  for(iterator it=m.map_.begin();it!=m.map_.end();++it)
  {
    if(map_.find(it->first)==map_.end())
      map_[it->first]=-it->second;
    else
    {
      map_[it->first]-=it->second;
      if(map_[it->first]==T(0) && !(it->first==Monom()))
        map_.erase(it->first);
    }
  }
  compute_hash();
  return *this;
}

template<typename T>
void Multinom<T>::compute_hash()
{
  hash_=0;
  for(typename MapType::const_iterator it=map_.begin();it!=map_.end();++it)
  {
    boost::hash_combine(hash_,it->first);
    boost::hash_combine(hash_,it->second);
  }
}

template<typename T>
inline Multinom<T> operator*(const Multinom<T> & mt1, const Multinom<T> & mt2)
{
  return Multinom<T>(mt1)*=mt2;
}

template<typename T>
inline Multinom<T> operator*(const Monom & m, const Multinom<T> & mt2)
{
  return Multinom<T>(mt2)*=m;
}

template<typename T>
inline Multinom<T> operator*(const Multinom<T> & mt2, const Monom & m)
{
  return Multinom<T>(mt2)*=m;
}

template<typename T>
inline Multinom<T> operator*(const T &t, const Monom & m)
{
  return Multinom<T>(m,t);
}

template<typename T>
inline Multinom<T> operator*(const T & t, const Multinom<T> & mt2)
{
  return Multinom<T>(mt2)*=t;
}

template<typename T>
inline Multinom<T> operator-(const Multinom<T> & mt1, const Multinom<T> & mt2)
{
  return Multinom<T>(mt1)-=mt2;
}

template<typename T>
inline Multinom<T> operator-(const Multinom<T> & mt1, const Monom & m2)
{
  return Multinom<T>(m2)-=mt1;
}

template<typename T>
inline Multinom<T> operator-(const Multinom<T> & mt1, const T & t)
{
  return Multinom<T>(t)-=mt1;
}

template<typename T>
inline Multinom<T> operator-(const Multinom<T> & m)
{
  return Multinom<T>(T(0))-=m;
}

template<typename T>
inline Multinom<T> operator+(const Multinom<T> & mt1, const Multinom<T> & mt2)
{
  return Multinom<T>(mt1)+=mt2;
}

template<typename T>
inline Multinom<T> operator+(const Multinom<T> & mt1, const Monom & m2)
{
  return Multinom<T>(m2)+=mt1;
}

template<typename T>
inline Multinom<T> operator+(const Multinom<T> & mt1, const T & t)
{
  return Multinom<T>(t)+=mt1;
}

template<typename T>
inline Multinom<T> operator^(const Multinom<T> & mt, int i)
{
  Multinom<T> res(T(1));
  for(int j=0;j<i;++j)
    res*=mt;
  return res;
}

}

#endif /* MULTINOM_H_ */
